/* **************************************************************************
   cbasicneuron.h  -  Basisklasse fr ein einfaches Neuron wie es in FFN verwendet wird
                             -------------------
    begin                : Tue Jun 6 2000
    copyright            : (C) 2000 by Jon Hennig
    email                : 
 ***************************************************************************/
/* **************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/


#ifndef CBASICNEURON_H
#define CBASICNEURON_H

#include <stl.h>

#include "tools.h" 		//for CChangingProperty, defines
class CChangingProperty;


#ifdef USE_LIBXML
	#include <libxml/tree.h>
#endif

#ifdef USE_CCOMMUNICATION
	#include "ccommunication.h"
#endif

class CDoubleVector;




/** Enum fr die verfgbaren Aktivierungsfunktionen.
    !ACHTUNG! Sollen neue Funktionen aufgenommen werden, so mssen diesem immer
    _hinten_ an die Liste angehangen werden, da die Aktivierungsfunktion als LONG
    gespeichert wird. Wird die bestehende Reihenfolge gendert, knnen alte Netze
    nicht mehr korrekt geladen werden.
    Fr jede neue Aktivierungsfunktion muss activation_function, derivative_activation_function
    und get_min_output, sowie get_max_output gendert werden.
    */
enum EActivationFunction
{   AF_LOGISTIC=0,        //die logistische Funktion
    AF_TANGENSH=1,        //Tangens Hyperbolicus
    AF_IDENT=2,           //Rckgabewert ist identisch der Aktivierung
    AF_NONE=3,            //keine Funktion aktiviert
    AF_LOGISTIC_SNNS=4    //die logistische Funktion, wie sie im SNNS definiert ist (als float -Funktion ohne Einstellung des slopes)
};

typedef double TNSynapticWeight;
typedef double TNPostSynapticValue;
typedef double TNPreSynapticValue;
typedef double TNActivationValue;

typedef double TNErrorValue;

typedef long TNTimeUnit;

class CNeuron;
class CNeuronicLayer;
class CSynapse;


/** Basisklasse fr die Implementierung eines Neurons.
    
    Diese Klasse selber ist nicht als anzuwendende Klasse gedacht, vielmehr bietet sie eine
    Basis fr die Implementation einer spezialisierten Klasse.
    
    Das Grundprinzip, das dieser Implementation zugrunde liegt, ist der lokale Charakter. Das
    heit die ganze Funktionalitt eines Algorithmus spielt sich auf der Ebene des Neurons ab.
    Idealerweise ist in den Schichten, bzw. im Netz keine zustzliche globale Information
    ntig, um einen bestimmten Algorithmus zu implementieren (Bsp. @ref CBackpropNeuron). Die
    Berechnung der Fehler, das Anpassen der Gewichte usw. erfolgt also nicht von "oben" sondern
    auf der Ebene des Neurons. Die hher liegenden Klassen (Schicht, Netz) steuern lediglich
    die Kommunikation der Neuronen untereinander.
    
    Es gibt allerdings Flle, in denen sich die Lokalitt nicht vollstndig wahren lsst. Zum
    Beispiel beim Ermitteln eines Gewinners einer Schicht von Neuronen (@ref CKohonenLayer).
    Bei der Implementation neuer Algorithmen, sollte trotzdem immer versucht werden, den lokalen
    Charakter so weit wie mglich zu erhalten.
    
    <b> TODO: </b>
        <li> bessere Untersttzung fr zeitabhngige Verarbeitung der Signale in den Neuronen
        <li> Funktion zum Einstellen der Aktivierungsfunktion (Untersttzung in process_cmd)
    </li>
    
    <b> BUGS: </b>
        <li> keine bekannt
    </li>
    
    @see CBackpropNeuron
    @see CInputNeuron
    @see CKohonenNeuron
    
    @author Jon Hennig
    */
class CNeuron
#ifdef USE_CCOMMUNICATION
:public CCommunication
#endif
{
  friend class CSynapse;
  friend class CNeuronicLayer;
public:

  CNeuron();

  /** Destruktor, entfernt alle Verbindungen zu dem Neuron
    */
  virtual ~CNeuron();

  //Zugriff auf Neuronen/Synapsen
  //================================
    /** liefert die InputSynapse mit der entsprechenden Nummer.
        @param number Nummer der Input-Synapse (beginnend bei 0)
        */
      CSynapse* input_synapse(long number);

  /** liefert die OutputSynapse mit der Nummer number.
  		@see input_synapse
  		*/
  CSynapse* output_synapse(long number);

  /** liefert das Input-Neuron mit der Nummer number.
  		Als Input-Neuron wird ein Neuron bezeichnet, das eine Ausgangsverbindung (Synapse)
  		zu dem aktuellen Neuron besitzt. Dieses muss in die Liste der verbundenen Neuronen aufgenommen
  		sein. (wird mit connect automatisch gemacht)
  		@param number Nummer des Input-Neurons (beginnend bei 0)
  		*/
  CNeuron* input_neuron(long number);

  /** liefert das Outputneuron mit der Nummer number
  		@see input_neuron
  		*/
  CNeuron* output_neuron(long number);

  /** liefert die Zahl der Input-Synapsen
  		*/
  virtual long get_anzahl_inputs();

  /** liefert die Zahl der Output-Synapsen
  		*/
  virtual long get_anzahl_outputs();

  //Ermitteln des Outputs  / Errors
  //===========================
  /** setzt den Output des Neurons, d.h. wendet die Aktivierungsfunktion auf die
      aktuelle Aktivierung an und speichert das Ergebnis in output. Die Aktivierung wird
      wieder auf 0 gesetzt, die alte Aktivierung als old_activation gespeichert
      */
  virtual TNPreSynapticValue calc_output();

  /** liefert den (letzten) Output des Neurons. (Dieser Wert ist in output gespeichert.)
  		*/
  virtual TNPreSynapticValue get_last_output();

  /** liefert den Fehler den das Neuron ausweist.
      Der Fehler ergibt sich aus der error_activation mal der Ableitung der activation (im Standardfall)
      Die Fehleraktivierung ergibt sich aus dem Vorhersagefehler (soll-ist) falls das Neuron in der Outputschicht ist.
      Und falls das Neuron in einer Hidden Layer ist aus der Summe von get_error_output()der Nachfolgeneuronen*Gewicht der Synapse zu den Nachfolgerneuronen (im Standardfall)
      Diese Funktion muss fr die meisten abgeleiteten Klassen berschrieben werden.
      */
  virtual TNErrorValue calc_error_output();

  /** liefert den letzten festgestellten error_output (ohne ihn neu zu ermitteln)
  		*/
  TNErrorValue get_last_error_output();

  //Verbindungen
  //============================
    /** verbindet das Neuron (Output) mit einem anderen Neuron (Input). Dabei wird eine Synapse mit dem entsprechenden
        Gewicht angelegt.
        @param weight das Gewicht der Verbindung
        */
  virtual void connect_output(CNeuron* other,TNSynapticWeight weight);

  //Weiterleitung
  //============================
  /** gibt den output des Neurons ber die Synapsen an alle Nachgeschalteten Neuronen weiter.
  		Dabei wird das Gewicht der Synapsen entsprechend jeweiligen Synapse beachtet.
  		@param time is ein Parameter um in Zukunft auch einen asynchronen, zeitgesteuerten Ablauf der
  								Ansteuerung der einzelnen Neuronen zu ermglichen. Zur Zeit wird er noch nicht verwendet.
  		*/
  virtual void feed_forward(TNTimeUnit time=-1);

  /** fhrt einen Feedbackschritt aus. Der Fehler wird unter Beachtung der Synapsen an
   	 alle vorgeschalteten Neuronen weitergeleitet
   	 @see feed_forward
   	 */
  virtual void feed_back(TNTimeUnit time=-1);

  //Setzen der Erregung/Einstellen von Werten
  //============================
  /** das Neuron wird zum Zeitpunkt time mit dem Wert excitation erregt
  		@param excitation Wert der zur Aktivierung des Neurons hinzugefgt werden soll.
  		@param time wird zur Zeit nicht beachtet. (siehe @ref feed_forward)
  	*/
  virtual void add_excitation(TNPostSynapticValue	excitation,TNTimeUnit time=0);

  /** fuegt der Error-aktivierung den entsprechenden Wert hinzu
    	die Fehleraktivierung entspricht der Aufsummation des zurckgeschickten Fehlers der Nachfolgerneuronen
    */
   virtual void add_error_activation(TNErrorValue error);

  /** setzt die Aktivierung des Neurons auf einen bestimmten Wert. Muss normalerweise nicht verwendet werden.
  		*/
  void set_activation(TNPostSynapticValue a_activation);

  /** setzt den Fehlerwert fr das Neuron, wenn moeglich
  		*/
  virtual void set_error(TNErrorValue);


  //Auslesen von Werten
  //============================
  /**  liefert den Vektor der Inputgewichte des Neurons
  		*/
  void get_weight_vector(CDoubleVector& vec);

  /** liefert die aktuelle Aktivierung des Neurons
  		*/
  TNPostSynapticValue get_activation();

  /** liefert den (letzten) Fehler des Neurons. D.h. der vor dem letzten Feedback-Schritt.
  		*/
  virtual TNErrorValue get_last_error();

  //Sonstiges
  //============================
  /** liefert den minimalen Wert, den die eingestellte Aktivierungsfunktion liefern kann. Wird fr die automatische
      Skalierung bentigt.
      */
  virtual double get_min_output();
  /** liefert den minimalen Wert, den die eingestellte Aktivierungsfunktion liefern kann.
      @see get_min_output
      */
  virtual double get_max_output();

  /** erzeugt eine Input-Synapse, die mit dem anderen Neuron verbunden ist.
      Kann in abgeleiteten Klassen berschrieben werden, wenn z.B. das Neuron
      eine andere als die Standardsynapse verwendet
      */
  virtual CSynapse* create_input_synapse(CNeuron* other,TNSynapticWeight weight);

  /** diese Variable wird nur bentigt, um den Typ dieses Neurons in add_neurons von CNeuronicLayer
      bestimmen zu knnen
      */
  static CNeuron* TYP;

  //Kommunikation
  //============================
#ifdef USE_CCOMMUNICATION
    /** Ausfhren eines Kommandos. Folgende stehen zur Verfgung: <b>
    		<li> "learning_rate","lr" ... bestimmt die Lernrate des Neurons (Startwert)
    		<li> "learning_rate_goal","lr_goal" ... bestimmt den Wert, auf den die Lernrate konvergieren soll
    		<li> "learning_rate_factor","lr_factor" ...  Der Faktor mit dem sich die Lernrate dem Konvergenzziel nhert. (blicherweise >0.99 und <1)
    		<li> "slope","sl" ...  Der Anstieg der Aktivierungsfunktion des Neurons
    		<li> "min_error","me" ... wenn der Fehler kleiner ist als min_error, werden auf 0 gesetzt. D.h. fr diese Werte erfolgt kein Lernen mehr.(Vermeidung von Overtraining)	  		</li>
    		<li> "output_weights","ow" ... Setzt die Gewichte der ausgehenden Synapsen auf einen zuflligen Wert zw. data[0] (min) und data[1] (max).(Standard -1,1)
    		</li>
    		@see CCommunication::process_cmd
      */
    virtual bool process_cmd(const char* command, const void* data,TPtrVector* add_data=0,long id=0);
#endif

  /** falls das die Funktion mit Argument true aufgerufen wird,
      wird das Lernen fr dieses Neuron erlaubt, ansonsten nicht.
      */
  void allow_learning(bool allow);
  /** falls das die Funktion mit Argument true aufgerufen wird,
      wird das Weiterleiten des Fehlersignals and die Vorgngerneuronen
      fr dieses Neuron erlaubt, ansonsten nicht.
      */
  void allow_feedback(bool allow);

  /** gibt zu einem Neuron dessen inputgewichte (als Zeilenvektor) aus (cout)
  		*/
  void print_input_vector();

  /** setzt den Output des Neurons auf den durch a_value bestimmten Wert
  		*/
  void set_output(double a_value);
  
#ifdef USE_LIBXML
  /** speichert in das XML-File, zu dem tree gehrt
  		*/
  virtual xmlNodePtr save_to_xml(xmlNodePtr tree);
  /** ldt aus dem XML-File, zu dem tree gehrt
  		*/
  virtual void load_from_xml(xmlNodePtr tree);
#endif

  /** setzt die Gewichte der Output-Synapsen auf einen zuflligen Wert zwischen  min und max
  		*/
  void set_output_weights(TNSynapticWeight min_weight=-1, TNSynapticWeight max_weight=1);

  /** Entfernt alle Verbindungen mit dem anderen Neuron, falls dieses angegeben ist.
      Ist kein anderes Neuron angegeben, werden alle Verbindungen gelscht.
      */
  void unconnect_all_outputs(CNeuron * other=0);
  /** Entfernt alle Verbindungen mit dem anderen Neuron, falls dieses angegeben ist.
  		Ist kein anderes Neuron angegeben, werden alle Verbindungen gelscht.
  		*/
  void unconnect_all_inputs(CNeuron * other=0);
  /** legt die Aktivierungsfunktion fr das Neuron fest. Achtung dies sollte normalerweise
      vor dem Trainieren des Netzes geschehen (beim Erstellen). Wenn fr eine Output-
      Schicht eine Skalierung gesetzt werden soll, so muss darauf geachtet werden,
      in welchem Bereich die Aktivierungsfunktion Werte liefert. Zum Beispiel 0..1 fr
      AF_LOGISTIC und -1..1 fr AF_TANH.
      @param set_default_params wenn dieser Wert true ist (default: false) werden die typischen
                                Lernparameter (z.Zt. nur learning_rate)
                                fr die Aktivierungsfunktion gesetzt.
                                (AF_TANH: 0.5, AF_LOGISTIC*: 0.9, sonst keine nderung)
      */
  void set_activation_function(EActivationFunction act_func, bool set_default_params=false);

  /** die verwendete Lernrate
      */
  CChangingProperty learning_rate;

  //Nur vorbergehend public zu Testzwecken
public:
    /** True, falls im Feedback-Schritt ein Weiterleiten des Fehlersignals an die Vorgngerneuronen
        erlaubt ist. Muss bei der Implementation einer abgeleiteten Klasse beachtet werden.
      	*/
    bool feedback_allowed;
protected:
  /** true, falls Lernen erlaubt ist, muss bei der Implementation einer abgeleiteten Klasse beachtet werden
  		*/
  bool learning_allowed;
  /** die aktuelle aktivierung
  		*/
  TNActivationValue activation;
  /** die letzte aktivierung
  		*/
  TNActivationValue old_activation;
  /** der letzte Output des Neurons (wird gleichzeitig mit get_output gesetzt)
  		*/
  TNPreSynapticValue output;
  /** Ist true, wenn der Output des Neurons per Hand gesetzt wurde.
  		In diesem Fall wird er im feed_forward-Schritt nicht erneut berechnet.
  		Wird von feed_forward immer auf false zurckgesetzt.
  		*/
  bool output_set;
  
public: //FIXME: sollte eigentlich protected sein (nur zum Testen)
  //Aktivierungsfunktion
  //====================
  /** die aktuell eingestellte Aktivierungsfunktion
  		*/
  EActivationFunction	act_activation_function;
  /** Steilheit der Aktivierungsfunktion - Standard 1
  		*/
  double slope;

  /* Die Schwelle wird standardmig auf 0 gesetzt und nicht modifiziert
     ber set_output wird sie in die Aktivierung mit einbezogen
     im Backpropneuron z.B. wird sie mit learn_step() geaendert
   	*/
  double schwelle;

  /** zur freien Verfgung
  		*/
  long tag;
  
protected:

  //Verbindungen
  /** die Verbindungen zu den Output-Neuronen
  		*/
  std::vector< CSynapse* > outputs;
  /** die Verbindungen zu den Input-Neuronen
  		*/
  std::vector< CSynapse* > inputs;

  //Funktionen
  /** informiert das Neuron das es ein neues Input aufnehmen soll
  		*/
  virtual void add_input(CSynapse* synapse);
  /** informiert das Neuron das es ein neues Output aufnehmen soll
  		*/
  virtual void add_output(CSynapse* synapse);

  /** Informiert das Neuron das es ein Input entfernen soll
  	  Entfernt nur den Verweis auf die Synapse, nicht die Synapse selbst
  	  */
  void remove_input(CSynapse* synapse);

  /** Informiert das Neuron das es ein Output entfernen soll
  	  Entfernt nur den Verweis auf die Synapse, nicht die Synapse selbst
  	  */
  void remove_output(CSynapse* synapse);

  //Aktivierungsfunktion
  //========================
  /** die Aktivierungsfunktion fr das Neuron
  		*/
  TNPostSynapticValue activation_function(TNActivationValue activation);

  /** Die Ableitung der Aktivierungsfunktion
  		*/
  double derivative_activation_function(double x);

   /** die Fehleraktivierung des Neurons -> ergibt Aufsummation des Fehlers
   		*/
  TNErrorValue error_activation;
  /** die Fehleraktivierung des Neurons (vor dem letzten Feedback-Schritt)
  		*/
  TNErrorValue old_error_activation;
  /** der ausgewiesene Fehler des Neurons (wird mit get_error_output ermittelt und abgefragt und
       mit get_last_error_output nur abgefragt
     	*/
  TNErrorValue error_output;
  /** der Fehler, der gerade noch als Fehler behandelt wird.
  	  alles, was vom Betrag her kleiner ist ergibt den Fehler 0 (um overfitting zu vermeiden
  	  und um Lernen in seltenen Gegenden des Fehlerraumes zu verbessern) - Standard:0
  	  */
  TNErrorValue min_error;

  
protected: 	
  /** Hilfsfunktion fr remove_input und remove_output
  		*/
  void remove_inoutput(CSynapse* synapse,std::vector<CSynapse*>& inoutputs);

};




//typedef CChangingProperty CProperty;





#endif


