// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//                          ccommunication.h  -  description
//                             -------------------
//    begin                : Tue Jan 23 2001
//    copyright            : (C) 2001 by Jon Hennig
//    email                : jon.hennig@gmx.net
// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// +                                                                         +
// +   This program is free software; you can redistribute it and/or modify  +
// +   it under the terms of the GNU General Public License as published by  +
// +   the Free Software Foundation; either version 2 of the License, or     +
// +   (at your option) any later version.                                   +
// +                                                                         +
// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++



#ifndef CCOMMUNICATION_H
#define CCOMMUNICATION_H

#include "tools.h"   //fr CString


/** TOOLS:
    ist als Basisklasse gedacht fr alle Klassen, die ber String-Befehle miteinander kommunizieren sollen.
    Sie stellt die Methode process_command(befehl) zur Verfgung. process_cmd muss in den abgeleiteten
    Klassen so implementiert werden, dass der befehl ausgefhrt wird. Wird der Befehl in der ab-
    geleiteten Klasse nicht erkannt, so muss er an die Basisklasse zur Verarbeitung weitergegeben
    werden. Weiterhin sind Befehle vorgesehen, die Befehle an bestimmte zu einem Objekt gehrenden
    Objekte weiterleiten. So wird "(x)" untersttzt um den Befehl an das x-te Element einer
    Auflistung weiterzuleiten. (*) leitet den Befehl an alle Elemente weiter.

    Sinn dieser Implementation ist es, dass fr alle von CCommunication abgeleiteten Klassen die
    selben Befehle benutzt werden knnen. Angewendet wird das um direkt interaktiv aus einer Umgebung
    Befehle eingeben zu knnen, die z.B. das Lernverhalten ndern.

    Das ermglicht Konstrukte der Art:
    <pre>
    double lr=0.9;
    net->process_command("layer(*).neuron(*).set_learning_rate",&lr);
    </pre>
    um an alle Neuronen des Netzes den Befehl zum Einstellen der Lernrate weiter zu geben.
    Neuronen, die den Befehl nicht kennen, ignorieren ihn.

    <b> TODO: </b>
    <li> weitere Parameter als (*) und (x) implementieren
    <li> Ausgabe der Informationen (infos) ist nicht gut implementiert -> besser stream, vielleicht einstellbar ber statische Variable?
    </li>

    <b> BUGS: </b>
    <li> keine bekannt
    </li>

    @author Jon Hennig
    */


 /** wird verwendet um der Funktion process_cmd zustzliche Daten zu bergeben.
  	 */
typedef std::vector<void*> TPtrVector; 		

class CCommunication {
public:

  CCommunication();
  virtual ~CCommunication();


  /** wie @ref process_cmd(command,data,..) nur dass data als Teil des Kommandos bergeben wird
  	  fhrt die Entschlsselung der bergebenen Parameter durch.
  	  So wird zum Beispiel aus process_command("slope=9") im Prinzip process_command("slope",9)
  	  Es kann entweder ein numerischer Parameter '=9' oder 2 numerische Parameter '=1:9'
  	  oder ein String ausgewertet werden: '="string'.
  	  Dabei werden 2 numerische Parameter a:b standardmig als whle Zufallswert zwischen a und b ausgewertet
    	*/
  bool process_command(const char* cmd);

  /** wie @see process_cmd(command,data) nur dass data eine double-Zahl ist.
    	*/
  bool process_command(const char* cmd, double data);

  /** stellt die zentrale Funktion zum Verarbeiten von Befehlen zur
      Verfgung. Muss (bei Bedarf) in den abgeleiteten Klassen berschrieben
      werden. command enthlt den Befehl (als String),data enthlt einen Zeiger auf
      die bergebenen Daten (id dient zur Identifikation von Antworten) Die berschreibende
      Klasse sollte am Ende die hnliche Funktion @process_cmd_parent der Parentklasse aufrufen.
      Falls true zurck geliefert wird, wurde erfolgreich ein Empfnger
      fr den Befehl gefunden. <p>
      @param data ist ein Zeiger auf die Daten, die der Funktion bergeben werden
      @param additional_data ist ein std::vector<void*>* auf zustzliche Daten, die der Fkt. bergeben werden.
                             Der Standardwert ist 0.
                             Die Position im Vektor gibt die Art der Daten an.
                             additional_data[0] ist ein Zeiger auf einen zustzlichen Datenwert. Wenn dieser
                             ungleich 0 ist, so wird als Wert fr die Funktion ein Zufallswert zwischen data*
                             und additional_data[0]* verwendet (fkt. nur bei Zahlenwerten)
      */
  virtual bool process_cmd(const char*  command, const void* data,TPtrVector* additional_data=0, long id=0);

  /** falls diese Variable auf wahr gesetzt ist werden Informationen der Klasse ausgegeben (mit @ref info),
      wenn nicht global_showinfo falsch ist.
      */
  bool showinfo;
  /** falls diese Variable auf falsch gesetzt ist werden in allen von CCommunication abgeleiteten Klassen keine
      Informationen mehr ber info ausgegeben. (Standard ist true)
      */
  static bool global_showinfo;
  /*  Die Folgenden Konstanten werden in Zusammenhang mit process_cmd bentigt:*/
  static const int ID_HELP=-1;            //Hilfe anzeigen
  static const int ID_NO_ERROR_MSG=-2;    //Keine Fehlermeldung, wenn Empfnger nicht gefunden

    
protected: // Protected methods
  /** Hilfsfunktion: gibt den bergebenen Text als Information (cout) aus (als Rckmeldung).
  		TODO: anders implementieren (streams!)
  		*/
  void infos(const CString& text);

  /** Hilfsfunktion: gibt die bergebene Zahl als Information (cout) aus (als Rckmeldung).
  		TODO: anders implementieren (streams!)
   		*/
  void infos(const double text);

  /** prft ob der Befehl command_to_check ein Befehl zur Weitergabe eines Befehls
  		an andere Objekte ist. Ermittelt, an welche Elemente der Befehl weitergegeben werden soll.
  		Start und stop geben das Anfangs- und Endelement zurck.In min und max  muss die Nummer des
  		kleinsten und des grten Elements bergeben werden. Dadurch sind Kostrukte der Art:
  	  <pre>
  		net->process_command("layer(*).neuron(*).set_learning_rate",0.9);
  		</pre>
  		einfach zu prfen. "(*)" liefert als min start und als max stop zurck. "(1)" liefert als min 1 und als max 1 zurck.
  		Denkbar, aber noch nicht implementiert ist z.B. "(5..7)".
  		@param command_to_check der zu prfende String
  		@param new_command liefert das weiterzugebende Kommando zurck.
  		@param min liefert das erste Element der Auflistung zurck
  		@param max liefert das letzte Element der Auflistung zurck
  		@param start muss bergeben werden gibt das erste Element einer Auflistung an. Wird zusammen mit
  		@param long das letzte Element der Auflistung (sonst wie start)
  		*/
  bool parse_command(const char * command_to_check,char*& new_command,long& start,long& stop,long min, long max);

  /** wird in process_cmd verwendet und prft, ob ein Befehl gltig ist.
      Falls die Funktion mit -1 als id aufgerufen wird, so wird die Beschreibung
      fr den Befehl zurckgeliefert.
      @param to_check der String, der geprft werden soll
      @param command gibt das zu prfende Kommando an
      @param description die Beschreibung fr den Befehl command
      @returns true, wenn to_check mit command bereinstimmt.
      */
  bool check_command(const char* to_check, const char * command, const char* description,long id=0);

  /** Hilfsfunktion, die zu data und add_data den Wert ermittelt, der verwendet werden soll (in process_cmd)
  		@param min_val Zeiger auf Variable in die minimal mgl. Wert geliefert wird (falls 0 wird dies nicht zurckgeliefert)
  		@param max_val analog min_val
  		*/
  double get_doub(const void* data,TPtrVector* add_data, double* min_val=0,double* max_val=0);
};

#endif
