// ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//                          cdoublevector.h  -  description
//                             -------------------
//    begin                : Tue Jun 20 2000
//    copyright            : (C) 2000 by Jon Hennig
//    email                :
// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// +                                                                         +
// +   This program is free software; you can redistribute it and/or modify  +
// +   it under the terms of the GNU General Public License as published by  +
// +   the Free Software Foundation; either version 2 of the License, or     +
// +   (at your option) any later version.                                   +
// +                                                                         +
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++


#ifndef CDOUBLEVECTOR_H
#define CDOUBLEVECTOR_H

#include <stl.h>
#include <std/bastring.h>

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
/** TOOLS: Ein Vektor von Doublewerten.

		Wird fast berall in dieser Bibliothek verwendet um auf einen Vektor von Double-Werten zuzugreifen.
		Zum Beispiel mssen die Input-Werte einer Neuronen-Schicht als Doublevektor bergeben werden.
		Die Handhabung ist beinah identisch mit double [x]. Allerdings erlaubt die Klasse einige Operationen
		auf dem Vektor. Sie ist von std::vector<double> der STL abgeleitet.
		
		Beipiel:
		<pre>
			CDoubleVector dv;
			dv.resize(3);		//3-Dim
			dv[0]=0; dv[1]=1; dv[2]=2;
			dv.add(1);
			cout << dv;		//ergibt 1, 2, 3
			dv.push_back(4);
			dv.push_back_doublevector(dv);
			cout << dv;		//ergibt: 1, 2, 3, 4, 1, 2, 3
		</pre>
		
		<b> TODO: </b>
	  <li> ev. weitere Operationen implementieren
	  <li> @ref distance_to fr verschiedene Skalarprodukte implementieren
	  <li> Zuweisung von normalen Vektor (double[] auf CDoubleVector
	  </li>
	
	  <b> BUGS: </b>
	  <li> keine bekannt
	  </li>
		
    @author Jon Hennig
  */
class CDoubleVector : public std::vector<double>  {
public: 
  CDoubleVector();

  /** generiert den Vektor auf Basis des Strings.
      Beispiel: "2.9 3.6 5.0"
      */
  CDoubleVector(basic_string<char> str,basic_string<char> delimiter=" ");
  ~CDoubleVector();

  /** skalare Subtraktion der Vektoren
  		*/
  friend CDoubleVector operator-(CDoubleVector& x,CDoubleVector& y);

  /** skalare Addition der Vektoren
  		*/
  friend CDoubleVector operator+(CDoubleVector& x,CDoubleVector& y);

  /** siehe @ref multiply(CDoubleVector)
  		*/
  friend CDoubleVector operator*(CDoubleVector& x,double l);

  /** liefert true, wenn jedes Element aus x kleiner als das entsprechende Element in y
  		*/
  friend bool operator<(CDoubleVector& x,CDoubleVector& y);

  /** Ausgabe auf den Stream.
  	*/
  friend ostream& operator<<(ostream& stream,CDoubleVector& x);
  /** Eingabe vom Stream.
  		*/
  friend istream& operator>>(istream& stream,CDoubleVector& x);

  /** multipliziert jedes Element des Vektors mit dem
      angegebenen Wert
      */
  void multiply(double value);
  /** siehe @ref multiply(double)
      */
  void operator*=(double value) {multiply(value);}
  
  
  /** dividiert jedes Element des Vektors durch den
  		angegebenen Wert
  		*/
  void divide(double value);
  /** siehe @ref divide(double)
      */
  void operator/=(double value) {divide(value);}

  /** verwandelt Werte, die zwischen scaled_min und scaled_max skaliert sind
  		wieder in die Originalwerte zurck.
  		*/
  void unscale(CDoubleVector* unscaled_min,CDoubleVector* unscaled_max,double scaled_min,double scaled_max);

  /** addiert zu dem Vector den anderen Vektor (skalar) hinzu
  		*/
  void add(CDoubleVector& other);
  /** siehe @ref add(other)
      */
  void operator+=(CDoubleVector& other) {add(other);}

  /** multipliziet den Vector mit dem anderen Vektor (skalar)
  		*/
  void multiply(CDoubleVector& other);

  /** addiert value zu jedem Element des Vektors hinzu
  		*/
  void add(double value);

  /** gibt den Vektor auf den Stream aus
  		*/
  ostream& write(ostream& stream,long width=0,char delimiter=' ');

  /** gibt den Vektor (von ..  bis)auf den Stream aus
  		*/
  ostream& write_partially(ostream& stream,long from, long to, long width=0,char delimiter=' ');

  /** hngt die Werte aus vec am Ende des Vektors an
  		*/
  void push_back_doublevector(const CDoubleVector &vec);

  /** errechnet die Distanz zu dem anderen Vektor
  	  @param other Der andere Vektor
  	  @param type Gibt an, nach welcher Norm der Abstand berechnet wird. (Wird zur Zeit ignoriert - immer Unendlich-Norm)
  	  */
  double distance_to(CDoubleVector &other, int type=0);

  /** liefert das Element des Vektors, der die minimale Differenz zu dem Wert (diff_value) besitzt
      */
  ulong elem_min_diff(double diff_value);
};



//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
/** TOOLS: Ein Feld (Vektor eines Vektors) von Doublewerten.

    entspricht im Wesentlichen  std::vector<CDoubleVector> und kann als Abkrzung
    dafr gesehen werden

    <b> TODO: </b>
    <li> wenn ntig weitere Funktionen definieren
    </li>

    <b> BUGS: </b>
    <li> keine bekannt
    </li>

    @author Jon Hennig
  */
class CDoubleVectorVector : public std::vector<CDoubleVector>  {
public:
    CDoubleVectorVector(){};

    /** liefert den Durchschnittsvektor zurck. (Mittelt ber jede Spalte)
        */
    CDoubleVector get_avg();
};





//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
/**	TOOLS: NUR_INTERN: implementiert eine Menge von long-Werten.
		(teilweise im Kohonen-Netz verwendet)			
  	
  	@author Jon Hennig
  */
class CLongSet : public std::set< long , less< long > >
{
public:
  /** entfernt alle Werte aus values_to_remove aus dem aktuellen Set
  		*/
  void remove(CLongSet& values_to_remove);
  /** fgt alle Werte aus values_to_add zu dem aktuellen Set hinzu
  		*/
  void add(CLongSet& values_to_add);
  /** gibt die Zahlen der Menge auf dem Stream aus.
  		*/
  friend ostream& operator<<(ostream& stream,CLongSet& x);

};



#endif
