/***************************************************************************
                          ckohonenlayer.cpp  -  description
                             -------------------
    begin                : Mon Feb 19 2001
    copyright            : (C) 2001 by Jon Hennig
    email                : jon.hennig@gmx.net
 ***************************************************************************/
/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/


#include "ckohonenlayer.h"

#include "ckohonenneuron.h"
#include "cdoublevector.h"
#include "math.h"
#include <set.h>

CKohonenLayer::CKohonenLayer()
{
	act_neighborhood_kernel=NK_CONE;
 	dimension=2;
 	breite=4;
 	tiefe=4;
 	
 	//Der Radius gibt an, in welchem Umkreis die Neuronen mitlernen drfen.
 	//Begonnen wird standardmig mit einem Radius von 6 als Grenzwert ist 1 eingestellt
 	//mit jedem Schritt nhert sich der Radius dem Zielwert um (ist-ziel)*Faktor
 	radius=6;
 	radius.set_goal(1);
 	radius.set_change_factor(0.999);
}

CKohonenLayer::~CKohonenLayer(){
}
/** liefert das Output der KohonenMap an die nachfolgenden Neuronen weiter */
void CKohonenLayer::feed_forward(TNTimeUnit time)
{
	assert(neuron(0));	//es muss mind. ein Neuron ex.
	
	//den Gewinner der einzelnen Neuronen ermitteln
	winner=get_winner();
	//und in den Neuronen setzen
	set_winner();
	
	
	for (long i=0;i<get_anzahl_neurons();i++)
	{	
		neuron(i)->feed_forward(time);
	}

}

/** setzt updated die Gewichte der Gewinner und der Nachbarn*/
void CKohonenLayer::feed_back(TNTimeUnit time)
{
	assert(neuron(winner));	
	
	//std::vector<long> neighbors;
	CLongSet neighbors;

	//als erstes mal das letzte Input bestimmen
	//Achtung, dass muss fr alle Neuronen der Kohonenlayer gleich sein
	CDoubleVector last_input;
	reinterpret_cast<CKohonenNeuron*>(neuron(0))->get_last_input(last_input);	
	//cout <<"Last Input :"<<last_input<<endl;
	
	//jetzt jedem Neuron entsprechen dem Abstand lernen lassen
	for (long i=0; i<get_anzahl_neurons();i++)
	{
		assert(dynamic_cast<CKohonenNeuron*>(neuron(i)));
		double learn_factor=neighborhood_kernel(get_abstand(i,winner),radius);
		if (learn_factor!=0)
		{
			//die lernrate wird in den jeweiligen Neuronen eingestellt
	  	reinterpret_cast<CKohonenNeuron*>(neuron(i))->learn_step(last_input,learn_factor);
	  }
	}
	radius.step(); //den radius entsprechend den Vorgaben verndert
}


/** liefert den Index des Winners der KohonenLayer*/
long CKohonenLayer::get_winner()
{
	double min_wert=10000000;
	long min=0;
	//das Neuron mit der niedrigsten Aktivierung ist der Gewinner, da es den niedrigsten Fehler hat
	for (long i=0;i<get_anzahl_neurons();i++)
	{	
		if (neuron(i)->get_activation()<min_wert)
		{ min=i;
			min_wert=neuron(i)->get_activation();
		}
	}	
	return min;
}

/** setzt is_winner beim Gewinner auf true, auf false bei den brigen*/
void CKohonenLayer::set_winner()
{
  CKohonenNeuron*	neur;
	for (long i=0;i<get_anzahl_neurons();i++)
	{	//FIXME: korrekte Typumwandlung??
		neur=reinterpret_cast<CKohonenNeuron*>(neuron(i));
		assert(neur);
		if (winner==i) neur->is_winner=true;
			else neur->is_winner=false;
	}	
}


/** liefert zu dem Neuron nr die spalte begonnen bei 0*/
long CKohonenLayer::get_spalte(long nr,long dim=2)
{
	assert(dim<=dimension && dim>0);
	switch (dimension)//die akt. eingestellte des Gitters
	{
		case 1: return nr;
		case 2:
			switch (dim) //die bergebene, zu der die Spalte ermittelt werden soll
			{
				case 1:	//in einem 2 Dim-Netz soll die Reihe bestimmt werden
								return nr/breite;
				case 2: //die Spalte soll bestimmt werden
								return nr%breite;
			}
		case 3:	//ein 3 dimensionales Gitter
			switch (dim) //die bergebene, zu der die Spalte ermittelt werden soll
			{
				case 1:	//in einem 3 Dim-Netz soll die Reihe bestimmt werden
								return (nr%(breite*tiefe))  /breite; //=rekursiv
				case 2: //die Spalte soll bestimmt werden
								return (nr%(breite*tiefe)) % breite;
				case 3: //die Tiefe soll bestimmt werden
								return nr/(breite*tiefe);
			}
  	 default: throw exception();
	}
	assert(false); //sollte niemals hierhin gelangen
	return 0;
}

/** liefert eine Zahl zwischen 0 und 1, die angibt, wie stark 
ein entsprechendes Neuron im Abstand abstand mitlernt */
double CKohonenLayer::neighborhood_kernel(double abstand, double max_radius)
{
	switch(act_neighborhood_kernel)
	{
		case NK_CYLINDER:	if (abstand<=max_radius) return 1.0; else return 0.0;
		case NK_CONE: 		if (abstand<max_radius) return 1.0-((double)abstand/(double)max_radius); else return 0.0;
		case NK_COS: 			if (abstand<max_radius) return cos(((double)abstand*M_PI_2/(double)max_radius)); else return 0.0;
		case NK_GAUSS1:		{	double zd=abstand/max_radius;
												if (abstand<max_radius) return exp(-(zd*zd)); else return 0.0;
											}
	  default :
	  	throw exception();
	}
}
/** liefert den Abstand zwischen den beiden Neuronen*/
double CKohonenLayer::get_abstand(long n1, long n2)
{
	double abstand=0;
	for (long i=1; i<=dimension;i++)
	{
		abstand=abstand+fabs(get_spalte(n1,i)-get_spalte(n2,i));
	}
	return abstand;

}

#ifdef USE_CCOMMUNICATION
/** ausfhren eines Kommandos*/
bool CKohonenLayer::process_cmd(const char* a_command, const void* data,TPtrVector* add_data=0,long id=0)
{
	if (check_command(a_command,"radius","bestimmt den Radius, um den andere Neuronen um das Gewinnerneuron mitlernen",id))
	{
		radius=get_doub(data,add_data);
		infos("Radius ist nun :");
		infos(radius);
		infos("\n");
		return true;
	}
	if (check_command(a_command,"radius_goal","bestimmt den Wert, auf den der Radius konvergieren soll.",id) )
	{
	  double wert=get_doub(data,add_data);
		radius.set_goal(wert);
		infos("Konvergenzziel fr den Radius ist nun :");
		infos(wert);
		infos("\n");
		return true;
	}
	if (check_command(a_command,"radius_factor","Der Faktor mit dem sich der Radius dem Konvergenzziel nhert. (blicherweise >0.99 und <1)",id) ||
	    check_command(a_command,"rad_factor","siehe radius_factor",id))
	{
	  double wert=get_doub(data,add_data);
		radius.set_change_factor(wert);
		infos("Der Faktor fr den Radius ist nun :");
		infos(wert);
		infos("\n");
		return true;
	}
	// die Parentklasse weis hoffentlich was damit anzufangen
	return CNeuronicLayer::process_cmd(a_command,data,add_data,id);
}
#endif

/** liefert die Nummer des Neurons, das beim letzten FeedForward-Schritt der Gewinner war */
long CKohonenLayer::get_last_winner()
{
	return winner;
}

#ifdef USE_LIBXML
//-----------------------------------------------------------------------------
/** speichert in das XML-File, zu dem tree gehrt */
xmlNodePtr CKohonenLayer::save_to_xml(xmlNodePtr tree){
	assert(tree);
	//FIXME: ungetestet
	//einen neuen Knoten anlegen (mit dem Klassennamen)
	xmlNodePtr subtree;
	subtree = xmlNewChild(tree, NULL, (const xmlChar *)"CKohonenLayer",NULL);
	xmlSetProp(subtree, (const xmlChar *)"type",(const xmlChar *) "layer");	//es handelt sich um ein Element vom Typ Layer
	//Zuerst die Basisklasse abspeichern
	CNeuronicLayer::save_to_xml(subtree);
	xmlSaveLong(subtree,"act_neighborhood_kernel",act_neighborhood_kernel);
	xmlSaveLong(subtree,"breite",breite);
	xmlSaveLong(subtree,"dimension",dimension);
	xmlSaveLong(subtree,"tiefe",tiefe);
	xmlSaveLong(subtree,"winner",winner);
	xmlSaveCProp(subtree,"radius",radius);
	return subtree;
}
//-----------------------------------------------------------------------------
/** ldt aus dem XML-File, zu dem tree gehrt */
void CKohonenLayer::load_from_xml(xmlNodePtr tree){
	assert(tree);
	//in den Knoten gehen (mit dem Klassennamen)
	if (strcmp((const char*)tree->name, "CKohonenLayer")) throw(666);
	xmlNodePtr subtree;
	subtree=tree->xmlChildrenNode;
	long i;
	while (subtree!=0)
	{
		if (!strcmp((const char*)subtree->name,"CNeuronicLayer")) CNeuronicLayer::load_from_xml(subtree);
  	if (xmlLoadLong(subtree,"act_neighborhood_kernel",i)) act_neighborhood_kernel=(ENeighborhoodKernel)i;
  	xmlLoadLong(subtree,"breite",breite);
  	xmlLoadLong(subtree,"dimension",dimension);
  	xmlLoadLong(subtree,"tiefe",tiefe);
  	xmlLoadLong(subtree,"winner",winner);
  	xmlLoadCProp(subtree,"radius",radius);
		subtree=subtree->next;
	}
}
#endif

