/* **************************************************************************
                          cneuronica.h  -  description
                             -------------------
    begin                : Fri Sep 21 2001
    copyright            : (C) 2001 by Jon Hennig
    email                : jon.hennig@gmx.net
 ***************************************************************************/
/* **************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/



#ifndef CNEURONICA_H
#define CNEURONICA_H

#include "global.h"
#include "tools.h"			//wegen CString, #defines

#ifdef USE_LIBXML
	#include <libxml/tree.h>
#endif


#include "cbackpropneuron.h"
#include "cstochasticbackpropneuron.h"
#include "cinputneuron.h"
#include "ckohonenneuron.h"
#include "cneuroniclayer.h"
#include "ckohonenlayer.h"
#include "csynapse.h"
#include "ckohonensynapse.h"
#include "cneuronicnet.h"
#include "cneuronicmultinet.h"
#include "cpatterns.h"
#include "cneuronicsnnsnet.h"

/** Datenstruktur zum Einlesen des Netzes (wird von nnd2net verwendet)
    */
struct TSynInfo{
    long pos;
    double weight;
};
/** Datenstruktur zum Einlesen des Netzes (wird von nnd2net verwendet)
    */
struct TNeuronInfo{
    long nr;
    double bias;
    double activation;
    EActivationFunction act_func;    
    std::vector< TSynInfo> weights;
};
typedef	 std::vector< std::vector< TNeuronInfo > >  TNeuronsVecVec;	


/** Klasse mit ausschlielich statischen Funktionen, die verschiedene Hilfsfunktionen zur Verfgung stellt.

    Bietet zum Beispiel Funktionen zum Erzeugen von Neuronen, basierend auf dem String des entsprechenden
    Klassennamen (@ref new_neuron). Dies wird u.a. bentigt um Netze aus eienem gespeicherten File laden zu
    knnen. Fr jede neue Klasse abgeleitet von CNeuron, CNeuronicLayer, CNeuronicNet, CSynapse, die aus
    einer Datei geladen werden soll, muss hier ein Eintrag ergnzt werden, da es mglich sein muss
    anhand des Strings (z.B. "CNeuronicNet" die entsprechende Klasse zu identifizieren und zu erzeugen.

    Weiterhin gibt es Funktionen zum Laden/Speichern von SNNS-Pattern-Files (@ref load_paterns,@ref save_patterns)

    @author Jon Hennig
    */

class CNeuronica{
private:
	 CNeuronica();
	~CNeuronica();
public:
	
	/** erzeugt ausgehend von neuron_name ein neues Neuron
			*/
	static CNeuron* new_neuron(const CString& neuron_name);
	
	/** erzeugt ausgehend von layer_name eine neue Layer
			*/
	static CNeuronicLayer* new_layer(const CString& layer_name);
	/** erzeugt ausgehend von layer_name und neuron_name eine neue
			Schicht mit anzahl Neuronen der entsprechenden Art
			*/
	static CNeuronicLayer* new_layer(const CString& layer_name, const CString& neuron_name, long nrof_neurons);
	
	/** erzeugt ausgehend von net_name eine neues Netz
			*/
	static CNeuronicNet* new_net(const CString& net_name);
	
	/** erzeugt ausgehend von synapsen_name eine neue Synapse und verbindet
	    diese mit den Neuronen
	  	*/
	static CSynapse* new_synapse(const CString& synapsen_name, CNeuron* pre_neuron,CNeuron* post_neuron);
	
	#ifdef USE_LIBXML
		/** erzeugt ausgehend von der Beschreibung in dem XML-tree ein neues Neuron
				*/
		static CNeuron* new_neuron(xmlNodePtr tree);
		/** erzeugt ausgehend von der Beschreibung in dem XML-tree eine neue Schicht
				*/
		static CNeuronicLayer* new_layer(xmlNodePtr tree);
		/** erzeugt ausgehend von der Beschreibung in dem XML-tree eine neues Netz
				*/
		static CNeuronicNet* new_net(xmlNodePtr tree);
		/** erzeugt ausgehend von der Beschreibung in dem XML-tree eine neue Synapse
		    und verbindet diese mit den Neuronen
		    */
		static CSynapse* new_synapse(xmlNodePtr tree,CNeuron* pre_neuron,CNeuron* post_neuron);
	#endif
		
	/** Erzeugt ein neues Netz entsprechend der Datei. Ldt das Netz in der Datei filename
			*/
	static CNeuronicNet* CNeuronica::load_new_net(const char* filename,  const char* format="DEFAULT");

  /** ldt aus einer Datei im SNNS-Pat-Format die In- und Outputs fr ein Netz.
  		*/
  static bool load_patterns(const char* snns_pat_filename, std::vector<CDoubleVector>& inputs, std::vector<CDoubleVector>& outputs);

  /** schreibt zusammengehrenden In-und Output in ein SNNS-Pattern-File
  		*/
  static bool save_pattern(ostream& out, CDoubleVector& input, CDoubleVector& output,long pos=0);

  /** schreibt die Muster in ein SNNS-Pattern-File
  		*/
  static bool save_patterns(const char* filename,std::vector< CDoubleVector > & inputs, std::vector<CDoubleVector> & outputs);

  /** verbindet net_out und net_in zu einem Netz. Die Parameter haben dieselben
			Bedeutungen wie bei CNeuronicMultiNet::connect().
			Liefert als Ergebnis die Nummer von Schicht layer_net_in in net_out zurck, -1,falls die Schicht gelscht wurde,
			weil sie keine Neuronen mehr enthlt
			@see CNeuronicMultiNet::connect
			*/
  static long merge( CNeuronicNet* net_out,  CNeuronicNet* net_in, ELayerCode layer_net_out, ELayerCode layer_net_in, long startneuron);

	#ifdef USE_LIBXML
    /** Konvertiert eine Datei im Neuronica-NND-Format ins SNNS-NET-Format.
    		Die Funktion basiert auf einer alten externen Implementierung, daher das etwas
    		eigenwillige Aussehen. Es wird (aus historischen Grnden) nicht das Netz erst aufgebaut,
    		sondern direkt die NND-(XML-)Datei geparst.<p>
        Bei einem Netz mit mehreren Input-Schichten, werden diese im SNNS-Netz
        durch unterschiedliche x-Parameter der entsprechenden Neuronen unterschieden
    		Es werden nicht alle Netztypen untersttzt (z.B. kein CNeuronicMultiNet)
    		*/
    static bool CNeuronica::nnd2net(const char* nndfile, const char * netfile);
	#endif

	//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
private:
//  /**	Datenstruktur zum Einlesen des Netzes (wird von nnd2net verwendet)
//  		*/
//  struct TSynInfo{
//  	long pos;
//  	double weight;
//  };
//  /**	Datenstruktur zum Einlesen des Netzes (wird von nnd2net verwendet)
//  		*/
//  struct TNeuronInfo{
//  	long nr;
//  	double bias;
//  	double activation;
//  	std::vector< TSynInfo> weights;
//  };
//  typedef	 std::vector< std::vector< TNeuronInfo > >  TNeuronsVecVec;	
	
	#ifdef USE_LIBXML
		/**	Hilfsfunktion (fr nnd2net) schreibt ein SNNS-File nach out Basierend auf dem geffneten NND-XML-Tree
				*/
		static bool nnd2net_intern(ostream& out,xmlNodePtr tree);

    /** Hilfsfunktion um die Informationen zu einem Neuron auszulesen (wird in nnd2net verwendet)
    		*/
    static TNeuronInfo nnd2net_get_neuron_info(xmlNodePtr node);
	#endif
	
  /** Hilfsfunktion zum Schreiben ins SNNS-Format
   		schreibt auf Basis der in neurons gespeicherten Werte das Netz
   		*/
  static void nnd2net_write_snns_file(TNeuronsVecVec& neurons,ulong nrof_input_layers,ostream& out);

  /** Hilfsfunktion zum Bestimmen der Nummer eines Neurons in einer TNeuronsVecVec - Struktur
  		*/
  static long nnd2net_get_neuron_nr(TNeuronsVecVec& neurons,long layer,long pos);

  /** Hilfsfunktion, liefert zu einer Aktivierungsfunktion den entsprechenden Namen im SNNS.
      Wenn default_name angegen ist, so wird, sollte das Ergebnis mit diesem bereinstimmen
      stattdessen "" als Ergebnis zurckgeliefert.
      */
  static CString get_snns_act_func_name(EActivationFunction act_func, const CString& default_name="");
};


#endif
