/* **************************************************************************
 cneuroniclayer.h  - Klasse, die mehrere Neuronen aufnimmt
                             -------------------
    begin                : Tue Jun 20 2000
    copyright            : (C) 2000 by Jon Hennig
    email                : 
 ***************************************************************************/
/* **************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/


#ifndef CNEURONICLAYER_H
#define CNEURONICLAYER_H

#include <stl.h>

#ifdef USE_LIBXML
	#include <libxml/tree.h>
#endif


#define USE_CCOMMUNICATION
#ifdef USE_CCOMMUNICATION
	#include "ccommunication.h"
#endif


#include "cbasicneuron.h"

class CDoubleVector;

enum EConnectionType
{
	CT_DEFAULT,								//die jeweilige Standardvorgabe verwenden
	CT_ALL_BEETWEEN_LAYERS,	 	//verbindet jedes Neuron der Schicht i mit jedem der Schicht i+1
	CT_ONE_TO_ONE          		//verbindet die Neuronen 1:1 das heisst alle Schichten mssen gleich viele Neuronen habe
};



/** Klasse zur Verwaltung mehrerer (=einer Schicht) Neuronen.

		Auf Grund des lokalen Charakters (siehe @ref CNeuron) der Implementation gengt es im Normalfall
		die einzelnen Neuronen ber diese Klasse zu steuern. Wenn der lokale Charakter verletzt wird, kann
		es ntig sein, spezialisierte Ableitungen dieser Klasse zu bilden (siehe @ref CKohonenLayer).
		
		Diese Klasse erlaubt es mehrere Neuronen gemeinsam anzusprechen (@ref feed_forward, @ref feed_back)
		und Werte zuzuweisen und auszulesen (@ref set_output, @ref get_output), sowie eine
		Skalierung/Normierung einzustellen (@ref set_scaling). Auerdem ist es mglich mehrere Schichten
		miteinander zu verbinden (@ref connect_output).
		
		Beispiel siehe @ref CNeuronicNet.
		
		<b> TODO: </b>
	  <li> Erweiterungen zum zeitbezogenen Ansteuern einzelner Neuronen (siehe auch CNeuron)
	  <li> Erweiterungen zum besseren Untersttzen von rekurrenten Netzen
	  </li>
	
	  <b> BUGS: </b>
	  <li> der Zufallszahlengenerator muss initialisiert werden (srand)
	  </li>

	  @see CNeuron
    @author Jon Hennig
  */
class CNeuronicLayer
#ifdef USE_CCOMMUNICATION
:public CCommunication
#endif
{
	friend class CNeuronicMultiNet;	//fr replace_subnet mit preserve_scaling
public: 
  /** verbindet alle Neuronen der Schicht mit dem Neuron neuron
      Dabei wird die Synapse auf einen zuflligen Wert zwischen min_weight und max_weight gesetzt
      */
  void connect_output(CNeuron* neuron, TNSynapticWeight min_weight, TNSynapticWeight max_weight);

  /** liefert die Anzahl der Neuronen in der Schicht
   		*/
  long get_anzahl_neurons();

  /** verbindet alle Neuronen dieser Schicht mit allen Neuronen aus der Schicht other.
      Dabei wird die Synapse auf einen zuflligen Wert zwischen min_weight und max_weight gesetzt.
      @param connection_type bestimmt die Art der Verbindung: CT_ALL_BEETWEEN_LAYERS verbindet jedes Neuron der einen Schicht mit jedem Neuron der anderen Schicht,
      																												CT_ONE_TO_ONE verbindet Neuron i der einen Schicht mit Neuron i der anderen Schicht.
      */
  void connect_output(CNeuronicLayer* other, TNSynapticWeight min_weight=-0.2, TNSynapticWeight max_weight=0.2, EConnectionType connection_type=CT_ALL_BEETWEEN_LAYERS);

  /** liefert das Neuron mit der Nummer number aus der Schicht.
  		Die Zhlung beginnt mit 0.
  		*/
  CNeuron* neuron(long number);

	CNeuronicLayer();
	virtual ~CNeuronicLayer();
	
  /** fgt anzahl Nuronen der Art T zur Schicht hinzu.
    	Die Funktion muss sich im Header befinden, da es sonst einen Fehler beim Kompilieren gibt.
    	*/
  template <class T> void add_neurons(T* mother,long anzahl)
	  { 	for (long i=1;i<=anzahl;i++)
		{  T* neuron = new T(); add_neuron(neuron);	}
	  }
	
  /** fgt das Neuron a_neuron zu der Menge der verwalteten Neuronen hinzu
			und liefert dessen Nummer zurck.
			*/
  long add_neuron(CNeuron* a_neuron);

  /** Fhrt einen Feed-Back-Schritt durch. Fr jedes Neuron der Schicht wird die Funktion
  		feed_back aufgerufen.
  		@see CNeuron::feed_back
  		*/
  virtual void feed_back(TNTimeUnit time=-1);

  /** Fhrt einen Feed-Forward-Schritt durch. Fr jedes Neuron der Schicht wird die Funktion
  		feed_forward aufgerufen.
  		@see CNeuron::feed_forward
  		*/
  virtual void feed_forward(TNTimeUnit time=-1);

  /** setzt die Aktivierungen der Neuronen in der Schicht.
  		@param activations ein Vektor mit den Aktivierungen. activations[i] entspricht der Aktivierung fr Neuron i.
  		Die Dimension des Vektors muss mit der Anzahl der Neuronen bereinstimmen.
  		@param use_scaling falls dieser Wert true ist, wird die voreingestellte Skalierung auf die Werte des Vektors angewendet
  		(siehe @ref set_scaling)
  		*/
  void set_activation(const CDoubleVector& activations, bool use_scaling=false);

  /** liefert die Outputs der Neuronen der Schicht im Vekor a_outputs zurueck. Dabei entspricht a_outputs[i] dem Output von Neuron i.
  		@param use_scaling falls dieser Wert true ist, wird die voreingestellte Skalierung auf die Werte des Vektors angewendet
  		(siehe @ref set_scaling).
  		*/
  void get_output(CDoubleVector& a_outputs, bool use_scaling=false);

  /** setzt den Fehler fr alle Neuronen der Schicht. Dabei entspricht a_errors[i] dem Fehler von Neuron i.
  		Die Dimension des Vektors muss mit der Anzahl der Neuronen bereinstimmen.
  		@param use_scaling falls dieser Wert true ist, wird die voreingestellte Skalierung auf die Werte des Vektors angewendet
  		(siehe @ref set_scaling)
  		*/
  void set_error(const CDoubleVector& a_errors, bool use_scaling=false);

  /** setzt die Skalierung so, dass der Input so eingerichtet wird, dass er zwischen scaled_min und scaled_max liegt
    	wenn der Output skaliert wird, muss der ursprngliche Wert zwischen scaled_min und scaled_max liegen,
    	damit er erfolgreich zurckskaliert werden kann.<p>
    	Beispiel:<p>
    	Fr ein Netz mit 2 Inputs (Temperatur 0..100, Volumen 200...2000) soll die Skalierung gesetzt werden:
    	Die skalierten Werte sollen zwischen -1 und 1 liegen:
    	<pre>
    	CDoubleVector smin(2),smax(2);
    	smin[0]=0; smax[0]=100;
    	smin[1]=200; smax[1]=2000;
    	input_layer->set_scaling(smin,smax,-1,1)
    	</pre>
    	Ein Aufruf von
    	<pre>
    	CDoubleVector bsp(2);
    	bsp[0]=0; bsp[1]=2000;
    	input_layer->set_output(bsp,true)
    	</pre>
    	stellt die output-Werte in den Neuronen wie gewollt auf -1 bzw. 1.
    	Analog liefert get_output(x,true) fr die in den Neuronen eingestellten Werte wieder die ursprnglichen Werte. 	
    	*/
  void set_scaling(CDoubleVector& a_unscaled_min, CDoubleVector& a_unscaled_max, double a_scaled_min, double a_scaled_max);

  /** setzt die Skalierung automatisch, so dass die skalierten Werte so liegen,
   	  dass automatisch der volle Bereich (-reserve) ausgeschpft wird. Dazu werden die Neuronen abgefragt,
   	  welchen Eingabe-Bereich sie untersttzen.
      @param reserve gibt den ungenutzten Randbereich an, der jeweils oben und unten am zu verfgbaren Bereich
                    eingespart wird. Angabe in Prozent. Wenn die Netzoutputs beispielsweise zwischen 0 und 1 liegen
                    gbe ein Wert reserve=10 einen Bereich von 0.1 bis 0.9
		  @see set_scaling
		  */
  void set_scaling_auto(CDoubleVector& a_unscaled_min, CDoubleVector& a_unscaled_max, double reserve=0);
	
	#ifdef USE_CCOMMUNICATION
	  /** Ausfhren eines Kommandos. Folgende stehen zur Verfgung: <b>
	  		<li> 	"neuron(i)" ... Gibt den Befehl an das Neuron i weiter ("neuron(*)" - an alle Neuronen der Schicht)
	  		</li>
	  		@see CCommunication::process_cmd
			  */
  	virtual bool process_cmd(const char* command, const void* data,TPtrVector* add_data=0,long id=0);
	#endif	

  /** ruft in allen Neuronen der Schicht die Funktion
			allow_learning auf. Falls true bergeben wird, wird
			das Lernen fr alle Neuronen der Schicht gestattet,
			sonst deaktiviert.
			*/
  void allow_learning(bool allow);

  /** ruft in allen Neuronen der Schicht die Funktion
      allow_feedback auf. Falls true bergeben wird, wird somit der Feedback-Schritt fr
      alle Neuronen der Schicht erlaubt, ansonsten verboten.
      */
  void allow_feedback(bool allow);

  /** gibt die Input-Gewichte der Schicht als Matrix aus (nach cout)
  		*/
  void print_input_matrix();

  /** gibt den Fehler der einzelnen Neuronen, die Fehleraktivierung, die Aktivierung und den Output aus
      */
  void print_output_and_error();

  /** setzt die Aktivierungen aller Neuronen der Schicht auf 0 zurck
  		*/
  void reset_activation();

  /** Setzt die Outputs der Neuronen in der Schicht auf die vorgegebenen
			Werte. Die Outputs der Neuronen werden in einem folgenden Feed-Forward-Schritt
			nicht (neu) berechnet.
			Dabei entspricht values[i] dem Output von Neuron i.
  		@param use_scaling falls dieser Wert true ist, wird die voreingestellte Skalierung auf die Werte des Vektors angewendet
  		(siehe @ref set_scaling).			
			*/
  void set_output(const CDoubleVector & values, bool use_scaling);

  #ifdef USE_LIBXML
	  /** ldt aus dem XML-File, zu dem tree gehrt
	  		*/
  	virtual void load_from_xml(xmlNodePtr tree);
	  /** speichert ins XML-File, an der Stelle tree
	  		*/
  	xmlNodePtr save_to_xml(xmlNodePtr tree);
  #endif

  /** liefert die letzten Fehler der Neuronen der Schicht in a_errors zurck. Dabei entspricht a_errors[i] dem Fehler von Neuron i.
  		@param use_scaling falls dieser Wert true ist, wird die voreingestellte Skalierung auf die Werte des Vektors angewendet
  		(siehe @ref set_scaling).			
			*/
  void get_error(CDoubleVector & a_errors, bool use_scaling);

  /** entfernt das Neuron mit der Nummer nrn_nr.
  		Achtung. Dadurch ndert sich die Nummerierung fr alle Neuronen mit einer Nummer grer als nrn_nr.
  		@param delete_neuron falls dieser Wert true ist, wird nach dem Entfernen aus der Schicht delete fr das Neuron aufgerufen (sonst nur entfernt)
  		*/
  void remove_neuron(long nrn_nr,bool delete_neuron=false);
	
protected:
	/** der Vektor mit den verwalteten Neuronen der Schicht
			*/
  std::vector< CNeuron *> neurons;

protected: // Protected attributes
  /** Die Faktoren mit denen jeder skalierte Wert multipliziert werden muss, um zu einen unskalierten Wert zu werden
  		Wird mit set_scaling bestimmt.
  		@see set_scaling
  		*/
  CDoubleVector* faktor;
  /** Offset, der nach dem Multiplizieren mit dem Faktor addiert werden muss, um einen unskal. Wert im gew. Bereich zu erzeugen
  		Wird mit set_scaling bestimmt.
  		@see set_scaling
  		*/
  CDoubleVector* offset;

};

#endif
