/* **************************************************************************
                          cneuronicnet.h  -  description
                             -------------------
    begin                : Sun Jan 7 2001
    copyright            : (C) 2001 by Jon Hennig
    email                : jon.hennig@gmx.net
 ***************************************************************************/
/* **************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/


#ifndef CNEURONICNET_H
#define CNEURONICNET_H

#include "global.h"
#include "tools.h"          //for #defines
#include "cpatterns.h"      //for CPattern

#ifdef USE_CCOMMUNICATION
	#include "ccommunication.h"
#endif

#include "cbasicneuron.h"

#ifdef USE_LIBXML
	#include <libxml/tree.h>
#endif

#include <map>								//for pos_info
#include "cneuroniclayer.h"		//for EConnectionType


class CNeuronicLayer;
class CDoubleVector;
class CPatterns;

/** Codes zum Ansprechen der In/Output-Schichten
		*/
enum ELayerCode
{
	OL0=-1,	//die erste Output-Layer (letzte Schicht)
	OL1=-2,	//die vorletzte Output-Layer
	OL2=-3,	//usw.
	OL3=-4,	
	OL4=-5,	
	OL5=-6,	
	IL0=0,	//die erste INput-Layer
	IL1=1,	//die zweite INput-Layer
	IL2=2,	//usw.
	IL3=3,
	IL4=4,
	IL5=5,
	IL6=6,
	IL7=7
};


/** NUR_INTERN: Struktur zum speichern von Informationen zur Position des Neurons im Netz
		*/
struct CNeuronPosInfo{
	long layer;
	long pos;
};


/** implementiert ein (knstliches) Neuronales Netz und ist gleichzeitig Basisklasse fr spezialisierte Netzklassen.

    Das Netz besteht aus mehreren Schichten (@ref CNeuronicLayer), welche es verwaltet.
    Die Funktionalitt wird normalerweise in  den Neuronen/Schichten festgelegt, so dass im Standardfall
    diese Basisklasse fr die Implementation eines Netzes verwendet werden kann.

    Beispiel fr die Anwendung:
    <pre>
    //-------------------------------------
    //Netz erzeugen
    CNeuronicNet net;
    //die Schichten anlegen und zum Netz hinzufgen
    net.add_layer(CInputNeuron::TYP,2);		//1. Schicht besteht aus 2 Inputneuronen
    net.add_layer(CBackpropNeuron::TYP,2);	// 2 Backpropneuronen in der verborgenen Schicht
    net.add_layer(CBackpropNeuron::TYP,1);	// 1 Backpropneuron in der Output-Schicht	
    //Neuronen des Netzes verbinden, zuflliges Gewicht zwischen -1 und 1
    net.connect(CT_ALL_BEETWEEN_LAYERS,-1,1);	  	
    //-------------------------------------
    //Training vorbereiten (Pattern laden)
    std::vector<CDoubleVector> inputs, outputs
    CNeuronica::load_patterns("xor.pat",inputs,outputs);
    //-------------------------------------
    //Das eigentliche Training
    CDoubleVector ergebnis;
    for (long j=0; j < 5000; j++)
    {
      for (ulong i=0; i < inputs.size(); i++ )
      {
      	//dem Netz die Inputs repsentieren
      	net->set_input(inputs[i],false);		//false=keine Skalierung verwenden
      	//die Werte vorwrts propagieren
      	net->feed_forward();
      	//das Netz-Ergebnis auslesen
      	net->get_output(ergebnis,false);		
      	//den Fehler ermitteln und setzen
      	net->set_error(outputs[i]-ergebnis,false);	
      	//den Fehler zurck propagieren
      	net->feed_back();
      }
    }
    //-------------------------------------
    //Anwendung des gelernten
    CDoubleVector in,out;
    in.push_back(0);
    in.push_back(1);
    net->calc_output(in,out,true);
    cout << "\nDas Ergebnis fr (" << in << ") lautet " << out;
    </pre>

    <b> TODO: </b>
    <li> Speichern/Laden im SNNS-NET-Format ermglichen (Code existiert teilweise schon [nnd2net])
    <li> Die Implementation der Klasse ist etwas Backprop-lastig. Muss im Zuge der Implementation anderer Netze ev. verbessert werden
    <li> Erweiterungen zum besseren Untersttzen von rekurrenten Netzen
    <li> Training bei @ref learn_all: Zufllige Anordnung der Patterns untersttzen
    <li> get_error LayerCodes verwenden lassen
    </li>

    <b> BUGS: </b>
    <li> nnd2net funktioniert nur fr ausgewhlte Netze
    </li>

    @see CNeuron
    @see CNeuronicLayer
    
    @author Jon Hennig
  */
class CNeuronicNet
#ifdef USE_CCOMMUNICATION
:public CCommunication
#endif
{
public:
	/** erzeugt ein leeres Netz
			*/
	CNeuronicNet();
	virtual ~CNeuronicNet();

  /** liefert einen Zeiger auf die Schicht nr zurck
  		*/
  CNeuronicLayer* layer(long nr);

  /** setzt den Fehler f. die letzte Schicht des Netzes.
  		Wird verwendet um z.B. einen Fehler zurck zu propagieren.
  		@param a_errors der Vektor mit den Fehlern
  		@param use_scaling falls true wird die eingestellte Skalierung verwendet (muss vorher definiert werden)
  		@see CNeuronicLayer::set_error
  		*/
  void set_error(const CDoubleVector& a_errors, bool use_scaling=false);

  /** liefert die Zahl der Schichten des Netzes
  		*/
  long get_anzahl_layers();

  /** liefert den Output der letzten Schicht des Netzes (das Netzergebnis).
  		@param a_outputs der Vektor mit den zurckgelieferten Outputs
  		@param use_scaling falls true wird die eingestellte Skalierung verwendet (muss vorher definiert werden)
  		@param nr bezeichnet die Schicht fr die der Befehl ausgefhrt wird. (OL0 bezeichnet die letzte Schicht, OL1 die vorletzte usw.) - Standard: OL0
  		@see CNeuronicLayer::get_output
  		*/
  void get_output(CDoubleVector& a_outputs, bool use_scaling=false,ELayerCode nr=OL0);

  /** setzt den Input fr die angegebene Schicht des Netzes. Bei dieser sollte es sich um eine Input-Schicht handeln.
  		Die Outputs der Input-Schicht werden auf die Werte in dem Vektor gesetzt.
  		@param input der Vektor Input-Werten
  		@param use_scaling falls true wird die eingestellte Skalierung verwendet (muss vorher definiert werden)
  		@param nr bezeichnet die Schicht - Standard 0 , die erste Schicht des Netzes
  		@see CNeuronicLayer::set_output
  		*/
  void set_input(const CDoubleVector& input, bool use_scaling=false, long nr_input_layer=0);

  /** fhrt einen Feed-Back-Schritt ber das gesamte Netz aus.
  		Dazu wird feed_back fr jede Schicht (gem ihrer Reihenfolge - beginnend mit der ersten) aufgerufen.
  		@see CNeuronicLayer::feed_back
  		*/
  virtual void feed_back(TNTimeUnit time=0);

  /** fhrt einen Feed-Forward-Schritt ber das gesamte Netz aus.
  		Dazu wird feed_forward fr jede Schicht (gem ihrer Reihenfolge - beginnend mit der letzten) aufgerufen.
  		@see CNeuronicLayer::feed_forward
  		*/
  virtual void feed_forward(TNTimeUnit time=0);

  /** fgt dem Netz eine Schicht Neuronen hinzu
    	@return die Nummer der Schicht im Netz
    	*/
  virtual long add_layer(CNeuronicLayer* a_layer);

  /** erstellt eine Schicht aus anzahl Neuronen der Art T und fgt diese zum Netz hinzu.
  		Das bergebene Neuron bestimmt nur den Typ des Neurons und stellt keine Kopiervorlage dar. (zur Verwendung mit CNeuron::TYP u. hnlichen)
    	@return die Nummer der Schicht im Netz
  		*/
  template <class T> long add_layer(T* mother,long anzahl)   //  Fehler, wenn in der cpp-Datei
	  { CNeuronicLayer* layer=new CNeuronicLayer();
	  	layer->add_neurons(mother,anzahl);
	  	return add_layer(layer);
	  }

  /** verbindet die des Netzes
      Dabei wird die Synapse auf einen zuflligen Wert zwischen min_weight und max_weight gesetzt.
      @param connection_type bestimmt die Art der Verbindung: CT_ALL_BEETWEEN_LAYERS verbindet jeweils jedes Neuron der einen Schicht mit jedem Neuron der anderen Schicht.
      																													Verbunden werden immer 2 benachbarte Schichten,und zwar so, dass die Output-Synapse der niederen die Input-Synapse der hheren Schicht ist.
      																												CT_ONE_TO_ONE verbindet jeweils Neuron i der einen Schicht mit Neuron i der anderen Schicht.
      																													Verbunden werden immer 2 benachbarte Schichten,und zwar so, dass die Output-Synapse der niederen die Input-Synapse der hheren Schicht ist.
      */
  virtual void connect(EConnectionType connection_type=CT_DEFAULT, TNSynapticWeight min_weight=-0.2, TNSynapticWeight max_weight=0.2);

  /** setzt das gesamte Netz zurck (lscht alle Schichten).
      @param force_delete falls gesetzt (Standard) werden alle Zeiger mit delete weggelscht
      */
  virtual void reset(bool force_delete=true);
	
	#ifdef USE_CCOMMUNICATION
	  /** Ausfhren eines Kommandos. Folgende stehen zur Verfgung: <b>
	  		<li> "layer(i)" ... Gibt den Befehl an die Schicht i weiter ("layer(*)" - an alle Schichten)
	  		<li> "print_weights" ... Gibt die Gewichte des Netzes aus (nach cout)
	  		<li> "save" ... speichert das Netz in eine Datei
	  		<li> "save_snns" ... speichert das Netz in eine SNNS-NET-Datei
	  		<li> "save_nnd" ... speichert das Netz in eine Datei im NND-Format
	  		<li> "load" ... ldt das Netz aus der angegebenen Datei
	  		</li>
	  		@see CCommunication::process_cmd
			  */
    virtual bool process_cmd(const char* command, const void* data,TPtrVector* add_data=0,long id=0);
	#endif	
	
  /** liefert zu dem Netzinput in in den entsprechenden Output nach out.
  		D.h. der Vektor in wird in die Input-Schicht des Netzes eingetragen.
  		feed_forward wird fr das Netz aufgerufen. Der Output der letzen Schicht wird
  		ausgelesen und in dem Vektor out zurckgegeben.
  		@param use_scaling falls true wird die eingestellte Skalierung verwendet (Skalierung muss sowohl fr Input- als auch fr die Output-Schicht eingestellt sein)
  		*/
  void calc_output(CDoubleVector& in, CDoubleVector& out,  bool use_scaling=false);

  /** liefert die Input-Matrizen der einzelnen Netz-Schichten nach cout.  	
  		*/
  void print_input_weights();
  /** gibt den Fehler der einzelnen Neuronen, die Fehleraktivierung, die Aktivierung und den Output aus
      */
  void print_output_and_error();

  #ifdef USE_LIBXML
	  /** ldt aus dem XML-File, zu dem tree gehrt, sollte normalerweise nicht direkt verwendet werden,
				sondern besser load()
				@see load
				*/
  	virtual void load_from_xml(xmlNodePtr tree);
	  /** speichert ins XML-File, an der Stelle tree, sollte normalerweise nicht direkt verwendet werden,
				sondern besser save()
				@see save
				*/
  	virtual xmlNodePtr save_to_xml(xmlNodePtr tree);
  #endif

  /** speichert in eine Datei im NND-Format bzw. im SNNS-NET-Format. Beim Speichern im SNNS-Format
      gibt die x-Position des Neurons die Schicht im Netz wider.
  		@param format "NND" steht fr das NND-Format, "SNNSNET" fr das SNNS-NET-Format
  		@return true, falls erfolgreich
      @see load
      @see CNeuronica::nnd2net
      */
 	virtual bool save(const char * filename, const char* format="DEFAULT");
	
	/** ldt das Netz aus einer Datei im XML-basierten NND-Format (Neuronica Network Definition)
			oder aus dem SNNS-NET-Format (zuknftig). Das direkte Laden aus dem SNNS-NET-Format ist zur
			Zeit noch nicht implementiert. Es existiert aber ein Tool nnd2net, welches eine Konversation
			von einem Format in das andere vornehmen kann. Eine Integration des entsprechenden Codes in
			die GPL-Version von Neuronica ist geplant.
			@param format "NND" steht fr das NND-Format, "SNNSNET" fr das SNNS-NET-Format
  		@return true, falls erfolgreich 	
  		*/
  virtual bool load(const char* filename , const char* format="DEFAULT");

  /** fhrt einen Lernschritt mit jedem der Patterns in CPatterns durch
  		*/
  void learn_all(CPatterns& patterns, bool use_scaling=false);

  /** liefert den Error der letzten Schicht des Netzes
  		@param a_errors der Vektor mit den zurckgelieferten Fehlern
  		@param use_scaling falls true wird die eingestellte Skalierung verwendet (muss vorher definiert werden)
  		@see CNeuronicLayer::get_error
  		*/
  void get_error(CDoubleVector & a_errors, bool use_scaling=false);

  /** ermittelt den summierten quadratischen Fehler des Neurons vor dem letzten Feed-Back-Schritt
  		- ungetestet -
  		*/
  double get_sse();

  /** liefert einen Zeiger auf die Input-Schicht nr. Die Schicht muss existieren.
  		Wenn mehr als eine Schicht eine Input-Schicht ist, mssen diese als Input-Schicht deklariert sein
  		(mit @ref set_noof_input_layers)
  		@see output_layer
  		*/
  virtual CNeuronicLayer * input_layer(long nr=0);

  /** liefert einen Zeiger auf die Output-Schicht nr.
      verwendet ELayerCode um die Outputschichten zu kennzeichnen
      Parameter OL0 liefert die letzte-Layer des Netzes,
      Parameter OL1 liefert die vorletzte usw.
      nr=IL0 liefert erste schicht des Netzes, IL1 die zweite usw.
  	  Geplant ist, einen Mechanismus zur Definition von In- und Outputschichten fr das Netz festzulegen
  	  */  	
  virtual CNeuronicLayer * output_layer(ELayerCode nr=OL0);

  /** ruft in allen Schichten des Netzes die Funktion
			allow_learning auf. Falls true bergeben wird., wird
			das Lernen fr alle Schichten des Netzes gestattet,
			sonst nicht.
			*/
  virtual void allow_learning(bool allow);

  /** entfernt die Schicht mit der Nummer aus der Verwaltung des NEtzes
			Schicht wird gelscht, wenn delete_layer true ist
			*/
  void remove_layer(long lay_nr,  bool delete_layer=false);

  /** entfernt die Schicht  aus der Verwaltung des NEtzes
			Schicht wird gelscht, wenn delete_layer true ist
			*/
  void remove_layer(CNeuronicLayer* lay, bool delete_layer=false);

  /** stellt die Lernrate aller Neuronen des Netzes auf den bergebenen Wert ein
  		*/
  virtual void set_learning_rate(double lr);

  /** erklrt die ersten noof_layers Schichten des Netzes zu Input-Schichten
  		*/
   void set_noof_input_layers(long nrof_layers);

  /** liefert die Anzahl der Input-Schichten fr ein Netz. Die Input-Schichten sind (zur Zeit) immer die ersten im Netzwerk.
  		*/
  long get_noof_input_layers();

  /** setzt min_error fr alle Neuronen der Output-Schicht.
      min_error ist der kleinste Fehlerwert, der noch als Fehler betrachtet wird (SNNS: dmax)
      */
  virtual void set_min_error(double min_err);

  /** Trainiert das Netz mit allen Patterns. Jeder Pattern wird noof_steps mal trainiert.
      */
  void train(CPatterns & patterns, long noof_steps, bool use_scaling=false);
  
  /** Trainiert das Netz mit den Inputs und den dazugehrigen Outputs. Jedes Input-Output-Paar wird
      noof_steps mal trainiert. (wie train mit Patterns)
      */
  void train(const std::vector< CDoubleVector>& inputs,const std::vector< CDoubleVector>& outputs, long noof_steps, bool use_scaling=false);
  
  /** Wenn true als Parameter bergeben wird, so wird bei jedem Aufruf von feed_back
      saved_patterns mit den aktuellen Werten von Input und Output gefllt.  Wird false
      bergeben, so wird nichts hinzugefgt.
      @param delete_old Wenn dieser Parameter true ist, werden alle schon gespeicherten
        Patterns in saved_patterns wieder gelscht (Standard: false)
      */
   void use_pattern_logging(bool use_it,bool delete_old=false);

protected:	// protected functions
  /** generiert Informationen zur Position des Neurons im Netz anhand
      seiner Adresse. Wird beim Speichern des Netzes bentigt.
      */
  void generate_pos_info();

  /** muss in jeder abgeleiteten Klasse vor dem Feedback-Schritt aufgerufen werden
      */
  void after_feed_back(TNTimeUnit time);

public:
	/** zustzliche Parameter fr bestimmte Netze (zur Zeit nur im SNNS-Netz verwendet)
			*/
	CDoubleVector params;

  /** Dieser Pattern wird bei jedem Aufruf von feed_back() gefllt, wenn pattern_logging==true ist
      */
  CPatterns logged_patterns;

protected: // Protected attributes
  /** die Schichten, aus denen das Netz besteht
  		*/
  std::vector<CNeuronicLayer*> layers;
  /** temp. Informationen zur Position eines Neurons im Netz anhand seiner Adresse
  		*/
  std::map< CNeuron*, CNeuronPosInfo > pos_info;

  /** intern zur Vermerkung der Anzahl der Input-Schichten
  		*/
  long noof_input_layers;

  /** siehe use_pattern_logging
      */
  bool pattern_logging;

};

#endif
