/***************************************************************************
                          tools.cpp  -  description
                             -------------------
    begin                : Mon Sep 3 2001
    copyright            : (C) 2001 by Jon Hennig
    email                : jon.hennig@gmx.net
 ***************************************************************************/
/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <assert.h>	
#include <stdlib.h>			//for rand
#include "global.h"
#include "tools.h"
#include <strstream.h>	//for ostrstream
#include <math.h>       //for log
#include <string>

#ifdef USE_LIBXML
	#include <libxml/parser.h>
#endif

//--------------------------------------------------------------------------
//| Allgemeine Funktionen                                                  |
//--------------------------------------------------------------------------


 /** liefert eine Zufallszahl zwischen min_value und max_value*/
extern double get_random_value(double min_value, double max_value){
   assert (max_value>=min_value);
   if (min_value==max_value) return min_value;
   double range=max_value-min_value;
   return ((double)rand()/(double)RAND_MAX) //liefert eine Zufallszahl zw. 0 und 1
     *range             //liefert eine Zahl zw. 0 und range
     +min_value;        //liefert eine Zahl zw. min_weight und min_weight+range

}

/** liefert eine Zufallszahl gem Gauss'scher Normalverteilung. Voraussetzung dafr ist
    eine Funktion get_random_value, die gleichverteilte Zufallszahlen liefert.
    Die Implementation beruht auf Donald E. Knuth's "The Art of Computer Programming". (siehe auch G. Marsaglia)
    @param mean       Mittelwert  - Standard: 0
    @param variance   Standardabweichung  - Standard: 1
    @param min_value  falls min_value oder max_value angegeben, so wird sichergestellt, dass das Ergebnis
                      immer im Bereich zwischen min und max liegt. Wenn nicht, wird ein gleichverteilter! Zufalls-
                      wert aus dem Bereich [min_value,max_value] ausgegeben. Fr min_value=MINDOUBLE und max_value=MAXDOUBLE
                      ergibt sich die (numerisch ber.) Gauss'sche Normalverteilung.
    @param max_value  siehe min_value, Standard: MAXDOUBLE
    */
extern double get_random_gauss(double mean=0, double variance=1,double min_value=-MAXDOUBLE, double max_value=MAXDOUBLE){
    //siehe auch JAVA-Klasse Random
    assert (max_value>=min_value);
    assert (max_value>=mean);
    assert (min_value<=mean);
    assert (variance>=0);
    if (min_value==max_value) return min_value;
    if (variance==0) return mean;
    double ret_val;
    static bool have_next_gauss_pre =false;  //bei einem Aufruf werden im Prinzip immer 2 Ergebnisse geliefert
    static double next_gauss_pre;            //wenn schon eines vorhanden, knnen wir Berechnung teilw. sparen

    if (have_next_gauss_pre){
        have_next_gauss_pre=false;
        ret_val=mean+sqrt(variance)*next_gauss_pre;   //wurde schon bei einem vorherigen Aufruf berechnet
    }else{
        double v1,v2,s;
        do{
            v1 = get_random_value(-1, 1);   //Bedingung: get_random_value muss gleichverteilten Wert liefern
            v2 = get_random_value(-1, 1);
            s = v1 * v1 + v2 * v2;
        } while (s >= 1);    //s muss fr die Berechnung >1 sein
        double multiplier = sqrt((-2.0*log(s))/s);
        next_gauss_pre = v2 * multiplier;
        have_next_gauss_pre = true;
        ret_val=mean+ sqrt(variance) * v1 * multiplier;
    }
    if ((ret_val<min_value) || (ret_val>max_value))
      //nicht im Bereich, eine gleichverteilte Zufallszahl des Bereiches wiedergeben
      return get_random_value(min_value,max_value);
    else
      //Zufallswert gem. Normalverteilung ausgeben
      return ret_val;
}




/** Erweiterung von throw*/
extern void throw_error(int err_number, const char* err_text,const char* function_name){
  if (function_name)
    cerr<<function_name<<": "<<err_text<<endl;
  else
    cerr<<err_text<<endl;
  throw(err_number);
}




//--------------------------------------------------------------------------
//| Stringfunktionen                                                       |
//--------------------------------------------------------------------------

//--------------------------------------------------------------------------
/**wie istream::getline(), aber fr CString*/
istream& getline(istream& strm, CString& str){
	//FIXME: Zeilenlnge nicht begrenzen
	char tmp_zeile[5000];
	istream& erg=strm.getline(tmp_zeile,5000);
  str.assign(tmp_zeile);
  return erg;
}


//--------------------------------------------------------------------------
/**konvertiert von einem Double zu einem String*/
extern CString double_to_string(double val){
		//Umwandeln
		ostrstream buffer;
		buffer<<val<<ends;	
		char * cstr=buffer.str();
		CString ergebnis(cstr);
		delete  cstr;
		return ergebnis;
}
//--------------------------------------------------------------------------
/**konvertiert von einem Long zu einem String*/
extern CString long_to_string(long val){
		ostrstream buffer;
		buffer<<val<<ends;	
		char * cstr=buffer.str();
		CString ergebnis(cstr);
		delete  cstr;
		return ergebnis;
}
//--------------------------------------------------------------------------
/**konvertiert von einem bool zu einem String*/
extern CString bool_to_string(bool val){
		CString ergebnis;
		if (val) ergebnis="true"; else ergebnis="false";
		return ergebnis;
}
//--------------------------------------------------------------------------
/**konvertiert von einem String zu einem Double*/
extern double string_to_double(const CString& val){
	return strtod(val.c_str(),0);
}
//--------------------------------------------------------------------------
/**konvertiert von einem String zu einem Double*/
extern float string_to_float(const CString& val){
		return atof(val.c_str());
}
//--------------------------------------------------------------------------
/**konvertiert von einem String zu einem Long*/
extern long string_to_long(const CString& val){
		return strtol(val.c_str(),0,10);
}
//--------------------------------------------------------------------------




//--------------------------------------------------------------------------
/** gibt wie printf aus, allerdings als CString
		*/
CString format_string(const char* format_str, ...) {
    CString ergebnis;
    char buffer[2048];
    va_list args;

    va_start (args, format_str);
    int i=vsprintf (buffer, format_str, args);
    va_end (args);
    assert(i<2048);	//sonst wurde ber den Buffern hinaus geschrieben

    ergebnis=buffer;
    return ergebnis;
}
//--------------------------------------------------------------------------
/** Fgt zwei Strings zusammen
    */
CString cat_strings(const char* str1, const char* str2){
    CString s1(str1);
    CString s2(str2);
    return (s1+s2);
}




//--------------------------------------------------------------------------
//| Funktionen fr die XML-Library                                         |
//--------------------------------------------------------------------------
#ifdef USE_LIBXML

//--------------------------------------------------------------------------
/** speichert einen Double-Wert in dem XML-File ab*/
extern void xmlSaveDouble(xmlNodePtr tree,const char * name,double value){
		assert(tree);
		assert(name);
    xmlNewChild(tree, NULL, (const xmlChar *)name, (const xmlChar *)double_to_string(value).c_str());
}
//--------------------------------------------------------------------------
/** speichert einen Long-Wert in dem XML-File ab*/
extern void xmlSaveLong(xmlNodePtr tree,const char * name,long value){
		assert(tree);
		assert(name);
    xmlNewChild(tree, NULL, (const xmlChar *)name, (const xmlChar *)long_to_string(value).c_str());
}
//--------------------------------------------------------------------------
/** speichert einen Long-Wert in dem XML-File ab*/
extern void xmlSaveBool(xmlNodePtr tree,const char * name,bool value){
		assert(tree);
		assert(name);
    xmlNewChild(tree, NULL, (const xmlChar *)name, (const xmlChar *) ((value)? "true":"false"));
}
//--------------------------------------------------------------------------
/** speichert einen CChangingProp-Wert in dem XML-File ab*/
extern void xmlSaveCProp(xmlNodePtr tree,const char * name,const CChangingProperty& value){
		assert(tree);
		assert(name);
		//einen neuen Knoten anlegen (mit dem Variablennamen) ...
		xmlNodePtr subtree;
		subtree = xmlNewChild(tree, NULL, (const xmlChar *)name,NULL);
		//...und die einzelnen Werte abspeichern
		xmlSaveDouble(subtree,"CProp_value",value.value);
		xmlSaveDouble(subtree,"CProp_goal",value.goal);
		xmlSaveDouble(subtree,"CProp_change_factor",value.change_factor);
}
//--------------------------------------------------------------------------
/** speichert einen CDoubleVector-Wert in dem XML-File ab*/
extern void xmlSaveDoubleVector(xmlNodePtr tree,const char * name,const CDoubleVector& vec){
		assert(tree);
		assert(name);
		//einen neuen Knoten anlegen (mit dem Variablennamen) ...
		xmlNodePtr subtree;
		subtree = xmlNewChild(tree, NULL, (const xmlChar *)name,NULL);
		//...und die einzelnen Werte abspeichern
		for (unsigned long i=0; i<vec.size();i++) xmlSaveDouble(subtree,"CDoubVec_value",vec[i]);
}
//--------------------------------------------------------------------------
/** ldt einen Double-Wert aus dem XML-File, gdw. der aktuelle Knotenname =name ist*/
extern bool xmlLoadDouble(xmlNodePtr tree,const char * name,double& value){
		assert(tree);
		assert(name);
	  char* cstr;
    if ((!strcmp((const char*)tree->name, name))){
	  	cstr = ( char*) xmlNodeListGetString(tree->doc, tree->xmlChildrenNode, 1);
	  	string str(cstr,strlen(cstr)+1); //string str(cstr); kopiert die abschlieende 0 nicht mit
		  value=string_to_double(str);	//Wert zuweisen
		  free(cstr);  //muss freigegeben werden, da xmlNodeListGetString Platz reserviert
		  return true;
		}//else
		return false;
}
//--------------------------------------------------------------------------
/** ldt einen Float-Wert aus dem XML-File, gdw. der aktuelle Knotenname =name ist*/
extern bool xmlLoadFloat(xmlNodePtr tree,const char * name,float& value){
		assert(tree);
		assert(name);
	  char* cstr;
    if (!strcmp((const char*)tree->name, name)){
	  	cstr = ( char*) xmlNodeListGetString(tree->doc, tree->xmlChildrenNode, 1);
	  	string str(cstr,strlen(cstr)+1); //string str(cstr); kopiert die abschlieende 0 nicht mit
		  value=string_to_float(str);	//Wert zuweisen
		  free(cstr);  //muss freigegeben werden, da xmlNodeListGetString Platz reserviert
		  return true;
		}//else
		return false;
}
//--------------------------------------------------------------------------
/** ldt einen Long-Wert aus dem XML-File, gdw. der aktuelle Knotenname =name ist*/
extern bool xmlLoadLong(xmlNodePtr tree,const char * name,long& value){
		assert(tree);
		assert(name);
	  char* cstr;
    if ((!strcmp((const char*)tree->name, name))){
	  	cstr = ( char*) xmlNodeListGetString(tree->doc, tree->xmlChildrenNode, 1);
	  	string str(cstr,strlen(cstr)+1); //string str(cstr); kopiert die abschlieende 0 nicht mit
		  value=string_to_long(str);	//Wert zuweisen
		  free(cstr);  //muss freigegeben werden, da xmlNodeListGetString Platz reserviert
		  return true;
		}//else
		return false;
}
//--------------------------------------------------------------------------
/** ldt einen Long-Wert aus dem XML-File, gdw. der aktuelle Knotenname =name ist*/
extern bool xmlLoadBool(xmlNodePtr tree,const char * name,bool& value){
		assert(tree);
		assert(name);
	  char* cstr;
    if ((!strcmp((const char*)tree->name, name))){
	  	cstr = ( char*) xmlNodeListGetString(tree->doc, tree->xmlChildrenNode, 1);
	  	value = (!strcmp("true",cstr));
		  free(cstr);  //muss freigegeben werden, da xmlNodeListGetString Platz reserviert
		  return true;
		}//else
		return false;
}
//--------------------------------------------------------------------------
/** ldt einen CProp-Wert aus dem XML-File, gdw. der aktuelle Knotenname =name ist*/
extern bool xmlLoadCProp(xmlNodePtr tree,const char * name,CChangingProperty& value){
		assert(tree);
		assert(name);
    if ((!strcmp((const char*)tree->name, name))){
    	//eine Ebene abtauchen
			xmlNodePtr subtree=tree->children;
			while (subtree)
			{
     		//...und die einzelnen Werte einlesen
     		xmlLoadFloat(subtree,"CProp_value",value.value);
     		xmlLoadFloat(subtree,"CProp_goal",value.goal);
     		xmlLoadFloat(subtree,"CProp_change_factor",value.change_factor);
     		subtree=subtree->next;
			}
		  return true;
		}//else
		return false;
}
//--------------------------------------------------------------------------
/** ldt einen CDoubleVector-Wert aus dem XML-File, gdw. der aktuelle Knotenname =name ist*/
extern bool xmlLoadDoubleVector(xmlNodePtr tree,const char * name,CDoubleVector& vec){
		assert(tree);
		assert(name);
    if ((!strcmp((const char*)tree->name, name))){	
    	//eine Ebene abtauchen
			xmlNodePtr subtree=tree->children;
			vec.resize(0);
			double doub;
			while (subtree)
			{
     		//...und die einzelnen Werte einlesen
     		if (xmlLoadDouble(subtree,"CDoubVec_value",doub)) vec.push_back(doub);
 	   		subtree=subtree->next;
			}
		  return true;
		}//else
		return false;
}
//--------------------------------------------------------------------------
extern bool xmlLoadInt(xmlNodePtr tree,const char * name,int& value){
		long l;
		bool erg=xmlLoadLong(tree,name,l);
		value=static_cast<int>(l);
		return erg;
}
//--------------------------------------------------------------------------
extern xmlNodePtr xmlSearchNode(xmlNodePtr tree, const char* name,bool depth_first){
	if (tree==0) return 0;
	xmlNodePtr erg;
	if ((!strcmp((const char*)tree->name, name))){
		//gefunden!!
		return tree;
	}else {
		//nicht gefunden
		if (depth_first){ //Falls Tiefensuche
	   	//eine Ebene abtauchen
			erg=xmlSearchNode(tree->children,name);
			if (erg) return erg;
			//else:     Einen Knoten weitergehen und dort probieren
			return xmlSearchNode(tree->next,name);
		}else{ //Falls Breitensuche
			//Einen Knoten weitergehen und dort probieren
			erg=xmlSearchNode(tree->next,name);
			if (erg) return erg;
	   	//else: eine Ebene abtauchen
			return xmlSearchNode(tree->children,name);
		}
	}
}


#endif


/****************************************************************************/
/* class CChangingProperty																									*/
/****************************************************************************/

/** Konstruktor */
CChangingProperty::CChangingProperty( )
{
	value=0;
	goal=0;
	change_factor=1;	//immer alles unverndert lassen
}
/** Konstruktor */
CChangingProperty::CChangingProperty(float a_value )
{
	value=a_value;
	goal=0;
	change_factor=1;	//immer alles unverndert lassen
}
/** liefert den Wert */
float CChangingProperty::get()
{
	return value+goal;	//value geht gegen 0
}
/** setzt den Wert
 */
void CChangingProperty::set(float a_value){
	value=a_value-goal;
}
/** setzt den Zielwert */
void CChangingProperty::set_goal(float a_value){
	float akt_wert=get();
	goal=a_value;
	//die verwendete Funktion konvergiert immer gegen 0 also den Wert entsprechend abndern
	set(akt_wert);
}
/** stellt den Faktor ein, mit dem sich die Eigenschaft
vom aktuellen Wert zum Zielwert hin entfernt (z.B. 0.999) */
void CChangingProperty::set_change_factor(float a_value){
	assert((change_factor>=0) &&(change_factor<=1) );	//was anderes wrde doch keinen Sinn machen
	change_factor=a_value;
}
/** macht einen Schritt Richtung Zielwert */
void CChangingProperty::step(){
	value=value*change_factor;
}


/****************************************************************************/
/* class CAvg                                                               */
/****************************************************************************/

/** Initialisierung*/
CAvg::CAvg(long a_noof_values){
    count=0;
    pos=0;
    sum=0;
    noof_values=a_noof_values;
    values.resize(noof_values,0);
}
/** fgt den Wert zu Durchschnittsberechnung hinzu*/
void CAvg::add(double value){
    count++;
    pos++; if (pos>=noof_values) pos=0;
    sum-=values[pos];
    sum+=value;
    values[pos]=value;
}
/** ermittelt den Durchschnitt fr die eingestellte Anzahl der Werte*/
double CAvg::get_avg(long a_noof_values){
    if (a_noof_values==-1) {
        //Standard
        if (count>noof_values) return sum/noof_values;
          else return sum/count;
    }else{
        //Wert ermitteln:
        throw_error(666,"Not implemented","CAvg::get_avg()");
        return 0; //dem Compiler zuliebe
    }
}
